package org.steveshipway.cheshirecat;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;
import org.steveshipway.cheshirecat.directive.*;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.ChunkCoordinates;
import net.minecraft.world.World;

public class Portfolio {
	// array of all blueprints
	HashMap<String, Blueprint> allBlueprints;
	HashMap<String, Blueprint> autoBlueprints;
	Logger logger;
	
	public Portfolio(Logger logger) {
		// initialise
		allBlueprints = new HashMap<String, Blueprint>();
		autoBlueprints = new HashMap<String, Blueprint>();
		this.logger = logger;
	}
	private Directive parse(String blueprint, String s, BufferedReader br) {
		Directive d = null;
		if( s.matches("(?i)\\s*call\\s(.*)") ) {
			d = new CallDirective(s,blueprint);
			return d;
		}
		if( s.matches("(?i)\\s*desc(.*)") ) {
			d = new DescDirective(s);
			return d;
		}
		if( s.matches("(?i)\\s*opt.*") ) {
			d = new OptionDirective(s);
			return d;
		}
		if( s.matches("(?i)\\s*block\\s.*") ) {
			d = new BlockDirective(s);
			return d;
		}
		if( s.matches("(?i)\\s*area\\s.*") ) {
			d = new AreaDirective(s);
			return d;
		}		// chance directive is odd as it requires the next item to be read
		if( s.matches("(?i)\\s*chance\\s+\\d+.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*chance\\s*(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		CheshireCat.logger.error("Unable to parse CHANCE directive: "+s);
	    		return null;
	    	}
	    	int c = Integer.parseInt(m.group(1));
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		CheshireCat.logger.error("Second directive NOT FOUND after CHANCE");
				return null;
			}
	    	Directive dd = parse(blueprint,ss,br);
	    	if( dd == null ) {
	    		CheshireCat.logger.error("Bad directive after CHANCE");
	    		return null;
	    	}
			d = new ChanceDirective(c,dd);
			return d;
		}
		// random directive requires next N directives to be read
		if( s.matches("(?i)\\s*random\\s+\\d+.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*random\\s*(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		CheshireCat.logger.error("Unable to parse RANDOM directive: "+s);
	    		return null;
	    	}
	    	int c = Integer.parseInt(m.group(1));
	    	String ss;
	    	List dl = new LinkedList();
	    	Directive dd = null;
	    	while( c > 0 ) {
	    		c -= 1;
	    		try {
	    			ss = br.readLine();
	    		} catch (IOException e) {
	    			CheshireCat.logger.error("Sufficient directives NOT FOUND after RANDOM");
	    			return null;
	    		}
	    		dd = parse(blueprint,ss,br);
	    		if( dd == null ) {
	    			CheshireCat.logger.error("Bad directive after RANDOM");
	    			return null;
	    		}
	    		dl.add(dd);
	    	}
			d = new RandomDirective(dl);
			return d;
		}
		if( s.matches("(?i)\\s*auto.*") ) {
			d = new AutoDirective(s);
			return d;
		}
		if( s.matches("(?i)\\s*(null|noop)\\s(.*)") ) {
			d = new NullDirective(s);
			return d;
		}
		return null;
	}
	private boolean loadFile( File f )  {
		// read in this File object
	    BufferedReader br;
	    String blueprint = null;
	    LinkedList steps = null;
	    String description = null;
	    Boolean visible = true;
	    int lineNo = 0;
		logger.info("["+f.getName()+"]: Parsing blueprint file ");
		try {
			br = new BufferedReader(new FileReader(f));
		} catch (FileNotFoundException e) {
			logger.error("["+f.getName()+"]: Unable to open file: "+e.getMessage());
			return false;
		}
	    try {
	        String line;
	        while ((line = br.readLine()) != null) {
	        	// work through lines in the file
	        	lineNo += 1;
	        	if( line.matches("\\s*#(.*)")) { continue; }
	        	if( line.matches("\\s*")) { continue; }
	        	if( line.matches("(?i)\\s*end(.*)")) {
	        		if( blueprint == null ) {
	        			logger.error("["+f.getName()+":"+lineNo+"]: Unbalanced END");
		        		visible = true;
		        		description = "No description";
		        		if( steps != null ) { steps.clear(); }
	        			continue;
	        		}
	        		// create new blueprint object
	        		Blueprint bp = new Blueprint(blueprint,description,visible,steps);
	        		allBlueprints.put(blueprint,bp );
	        		logger.info("["+f.getName()+":"+lineNo+"]: Created new blueprint name ["+blueprint+"]: "+description + " with "+steps.size()+" steps");
	        		if( bp.auto_chance > 0 ) {
	        			autoBlueprints.put(blueprint, bp);
	        		}
	        		// reset everything
	        		blueprint = null;
	        		visible = true;
	        		description = "No description";
	        		continue;
	        	} 
	        	Pattern p = Pattern.compile("^\\s*(public |private |)\\s*plan\\s+(\\w+)\\s*(.*)",Pattern.CASE_INSENSITIVE);
	        	Matcher m = p.matcher(line);
	        	if( m.find() ) {
	        		// a new definition!
	        		blueprint = m.group(2);
	        		description = m.group(3);
	        		if( m.group(1).matches("(?i)private.*")) {
	        			visible = false;
	        		}
	        		steps = new LinkedList();
	        		if( this.defines(blueprint) ) {
	        			logger.error("["+f.getName()+":"+lineNo+"]: Redeclaration of blueprint ["+blueprint+"] ");
	        			return false;
	        		}
	        		logger.info("["+f.getName()+":"+lineNo+"]: Starting new blueprint name ["+blueprint+"]");
	        		continue;
	        	}
	        	Directive d = parse(blueprint,line,br);
	        	if(d != null) {
	        		if(steps != null ) { steps.add(d); }
	        	} else {
		        	logger.error("["+f.getName()+":"+lineNo+"] Unknown or badly formatted directive in file: "+line);
		        	// return false;
	        	}
	        }
	    } catch (IOException e) {
			logger.error("["+f.getName()+"]: Unable to read file: "+e.getMessage());
			return false;
		} finally {
	        try {
				br.close();
			} catch (IOException e) {
				// ignore
			}
	    }
	    if( blueprint != null ) {
	    	logger.warn("["+f.getName()+":"+lineNo+"]: Incomplete blueprint declaration ["+blueprint+"] ");
	    	return false;
	    }
		return true;
	}
	public void load(String d) {
		// load all blueprints in from config directory d
		logger.info("Initialising portfolio from " + d);
		allBlueprints.clear();
		File folder = new File(d);
		for (final File fileEntry : folder.listFiles()) {
		    if (! fileEntry.isDirectory() && fileEntry.getName().endsWith(".cfg")) {
		        if( ! loadFile( fileEntry ) ) {
		    	    logger.error("["+fileEntry.getName()+"]: Unable to read blueprint file " + fileEntry.getPath());
		        }
		    }
		}
	    logger.info("There are " + allBlueprints.size() + " blueprints defined");
	    logger.info("There are " + autoBlueprints.size() + " auto-generate blueprints defined");
	}
	// save the blueprint to the specified file
	public boolean save(String s) {
		boolean rv = true;
		if( s == null ) { return false; }
		Blueprint bp = allBlueprints.get(s);
		if( bp == null ) { return false; }
		String f = CheshireCat.bppath + File.separator + s + ".cfg";
		Writer writer = null;
		try {
		    writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream(f), "utf-8"));
		    writer.write("# Blueprint for: "+bp.name+"\n");
			writer.write(bp.asText()+"\n");
		} catch (IOException ex) {
			logger.error("Unable to write file "+f+": "+ex.getMessage());
			rv = false;
		} finally {
		    try {writer.close();} catch (Exception ex) {}
		}
		return rv;
	}
	// check to see if the blueprint exists
	public boolean defines(String s) {
		if( s == null ) { return false; }
		return allBlueprints.containsKey(s);	
	}
	// fetch the blueprint
	public Blueprint blueprint(String s) {
		Blueprint bp = allBlueprints.get(s);
		return bp;
	}
	// build the blueprint
	public boolean build(World world, String s,ChunkCoordinates playerLocation, int r) {
		if( s == null ) {
			logger.error("Somehow received null blueprint name.  How did you call this?");
			return false;
		}
		if( world.isRemote ) {
			logger.error("Portfolio build called when World is remote?  Shouldnt happen.");
			return false;
		}
		if( playerLocation == null ) {
			logger.error("Somehow received null player location.  That shouldn't be possible.");
			return false;
		}
		logger.info("Calling build for blueprint " + s + " at " + playerLocation.posX + "," + playerLocation.posY + "," + playerLocation.posZ + " rotation " + r);
		if( ! this.defines(s) ) {
			logger.warn( "Invalid blueprint name " + s);
			return false;
		}
		boolean rv = allBlueprints.get(s).build(world,playerLocation,r&3);
		return rv;
	}
	public List list(String s) {
		List l = new ArrayList();
		Set set = allBlueprints.entrySet();
	    Iterator i = set.iterator();
	    l.clear();
	    while(i.hasNext()) {
	    	Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( bp.isVisible() && bp.name.startsWith(s)) {
	        	l.add(bp.name);	      
	    	}
	    }
		return l;
	}
	public List list() {
		return this.list("");
	}
	public Iterator iterator() {
		return this.allBlueprints.keySet().iterator();
	}
	public Iterator auto_iterator() {
		return this.autoBlueprints.keySet().iterator();
	}
	public List help() {
		List l = new ArrayList();
		Set set = allBlueprints.entrySet();
	    Iterator i = set.iterator();
	    int c = 0;
	    while( i.hasNext()) {
            Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( bp.isVisible() ) { c += 1; }
	    }
	    i = set.iterator();
	    l.add("There are " + allBlueprints.size() + " blueprints defined ("+c+" available)");
	    while(i.hasNext()) {
	    	String s = new String();
            Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( bp.isVisible() ) {
              s = bp.name + " : " + bp.helpText();
	          l.add(s);	         
            }
	    }
		return l;
	}
	public boolean delete(String s) {
		if( ! defines(s) ) { return true; }
		allBlueprints.remove(s);
		return true;
	}
	// copy from the world
	public Blueprint copy(World world, String name, ChunkCoordinates from, ChunkCoordinates to, int opts ) {
		Blueprint bp = null;
		LinkedList steps = new LinkedList();
		if( defines(name) ) {
			logger.error("Blueprint "+name+" already exists");
			return null;
		}
		// iterate over the entire area.  Make sure coordinates are in correct order.
		if( from.posX > to.posX ) {
			int q = from.posX; from.posX = to.posX; to.posX = q;
		}
		if( from.posY > to.posY ) {
			int q = from.posY; from.posY = to.posY; to.posY = q;
		}
		if( from.posZ > to.posZ ) {
			int q = from.posZ; from.posZ = to.posZ; to.posZ = q;
		}
		int curDv;
		Block curBlock;
		ChunkCoordinates offset = null;
		NBTTagCompound nbt = null;
		TileEntity te = null;
		String nbtStr;
		for( int x = from.posX; x <= to.posX; x += 1 ) {
			for( int y = from.posY; y <= to.posY; y += 1 ) {
				for( int z = from.posZ; z <= to.posZ; z += 1 ) {
					curBlock = world.getBlock(x,y,z);
					if( curBlock != Blocks.air && curBlock != Blocks.water && curBlock != Blocks.lava && curBlock != Blocks.flowing_water && curBlock != Blocks.flowing_lava ) {
						curDv = world.getBlockMetadata(x,y,z);
						offset = new ChunkCoordinates();
						offset.posX = x - from.posX; 
						offset.posY = y - from.posY; 
						offset.posZ = z - from.posZ;
						te = world.getTileEntity(x,y,z);
						if(te != null) {
						//Block b = te.getBlockType();
							nbt = new NBTTagCompound();
							te.writeToNBT(nbt); // note this does not remove existing tags
							nbt.removeTag("x"); // remove the tags we don't need
							nbt.removeTag("y");
							nbt.removeTag("z");
							nbtStr = nbt.toString();							
						} else {
							nbtStr = null;
						}
						steps.add(new BlockDirective(curBlock,curDv,false,offset,nbtStr));
						//logger.info("Added block of type "+Block.blockRegistry.getNameForObject(curBlock)+" at position "+x+","+y+","+z+" ("+offset.posX+","+offset.posY+","+offset.posZ+")");
					} else {
						//logger.info("Omitting block of type "+Block.blockRegistry.getNameForObject(curBlock)+" at position "+x+","+y+","+z);
					}
				}
			}
		}
		
		// create blueprint object
		bp = new Blueprint(name,"",true,steps);
		allBlueprints.put(name, bp);
		return bp;
	}
}
